// C Source File
// Created 2/5/2002; 6:58:02 PM
#define MIN_AMS 200


#define USE_TI89              // Produce .89z File

#define OPTIMIZE_ROM_CALLS    // Use ROM Call Optimization

#define SAVE_SCREEN           // Save/Restore LCD Contents

#define EDITOR			//comment this out to compile without the editor, it's about 3 Kbytes smaller

#define headup 0
#define bodyvert 16
#define upleft 32
#define tailup 48
#define food 64

#define bodyhorz 80

#define upright 96
#define downright 112
#define downleft 128

#define headright 144
#define headdown 160
#define headleft 176

#define tailright 192
#define taildown 208
#define tailleft 224

#include <tigcclib.h>         // Include All Header Files

#include "extgraph.h"


//These are little macros I wrote for position checking, used in the AI and in DrawSnake() for deciding which body sprites to use
#define IsAbove(i1, i2) ((body[i1].x == body[i2].x) && (body[i1].y+1 == body[i2].y))
#define IsBelow(i1, i2) ((body[i1].x == body[i2].x) && (body[i1].y-1 == body[i2].y))
#define IsLeft(i1, i2) ((body[i1].x+1 == body[i2].x) && (body[i1].y == body[i2].y))
#define IsRight(i1, i2) ((body[i1].x-1 == body[i2].x) && (body[i1].y == body[i2].y))

//Binary manipulation macros I wrote for the sprite editor
#define BitOn(var, b) var = var | (1<<(7-b))
#define BitOff(var, b) var = var & ~(1<<(7-b))
//#define BitOff(var, b) var = var & ( ^ (1<<(7-b)))


#define GraySprite(x, y, s) \
	Sprite8(x, y, 8, &sprites[s], lbuffer, SPRT_OR); \
	Sprite8(x, y, 8, &sprites[s+8], dbuffer, SPRT_OR);


void *lbuffer = NULL, *dbuffer = NULL;
FILE *cfgfile = NULL;
INT_HANDLER OldInt5 = NULL;
short font;
static unsigned char sprtdata[]={
	0x18, 0x24, 0x42, 0xA5, 0x81, 0x99, 0x99, 0x99,		//Headup
	0x18, 0x3C, 0x7E, 0xFF, 0xFF, 0xE7, 0xE7, 0xE7,
	0x99, 0x99, 0x99, 0x99, 0x99, 0x99, 0x99, 0x99,		//Body
	0xE7, 0xE7, 0xE7, 0xE7, 0xE7, 0xE7, 0xE7, 0xE7,
	0x99, 0x19, 0x19, 0xF9, 0xF9, 0x01, 0x01, 0xFF,		//Upleft
	0xE7, 0xE7, 0xE7, 0x07, 0x07, 0xFF, 0xFF, 0xFF,
	0x18, 0x24, 0x42, 0x99, 0x99, 0x99, 0x99, 0x99,	  //Tailup
	0x18, 0x3C, 0x7E, 0xE7, 0xE7, 0xE7, 0xE7, 0xE7,
	0x00, 0x00, 0x3C, 0x7E, 0xFF, 0xFF, 0x7E, 0x3C,		//Food
	0x04, 0x08, 0x18, 0x7E, 0xFF, 0xFF, 0x7E, 0x3C
};

unsigned char *sprites = NULL;

int ai, aiused, dir, i, score, hscore, homequit, key, c;

struct FOOD {
	int x, y;
} Food;


struct BODY {
	int x, y;
} *body = NULL;

int length;
char fps;
volatile int tmr = 0;
//char cbuffer[50];

// NOTE: If you see a somewhat random comment, it's probably something I just put in the code to help myself, 
// such as a snippet from the documentation or some website.  

//All other comments are there to explain my code. 

//Feel free to use any of my code, just give me credit if you do.  


//these two archive/unarchive the config file
void Archive(void) {
	if (!DerefSym(SymFind(SYMSTR("bnibcfg")))->flags.bits.archived) 
		EM_moveSymToExtMem(SYMSTR("bnibcfg"), HS_NULL);
}
void UnArchive(void) {
	if (DerefSym(SymFind(SYMSTR("bnibcfg")))->flags.bits.archived) 
		EM_moveSymFromExtMem(SYMSTR("bnibcfg"), HS_NULL);
}

//Draws a centered string...
void DrawStrCenter(char *str, short y) {
	DrawStr((160-DrawStrWidth(str, FontGetSys()))/2, y, str, A_NORMAL);
}

//Rotates an 8x8 sprite 90 degrees clockwise.
void RotateSprites() {
	int x = 0, n = 0, sprt = 0, rsprt = 0;
	int sprtn[] = {bodyvert, upleft, upright, downright, headup, headright, headdown, tailup, tailright, taildown};
	memset(&sprites[80], 0, 160);
	for (i = 0; i <= 9; i++) {
		sprt = sprtn[i];
		rsprt = 80+(16*i)/*rsprtn[i]*/;
		for (n = 0; n <=7; n++) {
			for (x = 0; x <=7; x++) {
				if ((sprites[sprt+n] & (1<<(7-x))) == (1<<(7-x)))
					sprites[rsprt+x] = (sprites[rsprt+x] | (1<<(n)));
				if ((sprites[sprt+n+8] & (1<<(7-x))) == (1<<(7-x)))
					sprites[rsprt+x+8] = (sprites[rsprt+x+8] | (1<<(n)));
			}
		}
	}
}

DEFINE_INT_HANDLER(CustomInt5) {
	tmr = 1;
	ExecuteHandler(OldInt5);
}


CALLBACK void ExitFunc(void) {
	if (homequit) goto exitquick;

	GrayOff();

	clrscr();
	SetIntVec(AUTO_INT_5, OldInt5);
	//This restores the timer interrupt back to normal for a HW2 calc.  If you know a more elegant or effective way, please email me
	pokeIO(0x600017, 204);
	

	UnArchive();
	
	cfgfile = NULL;
	cfgfile = fopen("bnibcfg", "r+b");
	if (cfgfile) {
		fwrite(&fps, 1, 1, cfgfile);
		fclose(cfgfile);
	}
	
	//Handle high scores, most of the function names are self-explanatory
	cfgfile = NULL;
	cfgfile = fopen("bnibcfg", "rb");
	if (cfgfile) {
		fseek(cfgfile, 1, SEEK_CUR);
		fread(&hscore, 2, 1,cfgfile);
		fclose(cfgfile);
	}
	cfgfile = NULL;
	printf("Your score was: %d\n\n", score);
	if (score > hscore && !aiused) {		
		cfgfile = fopen("bnibcfg", "r+b");
		if (cfgfile) {
			fseek(cfgfile, 1, SEEK_CUR);
			fwrite(&score, 2, 1, cfgfile);
			fclose(cfgfile);
			printf("New high score!\n\n");
			printf("The old high score: %d\n", hscore);
		}
	}	
	else printf("High score: %d\n", hscore);
	if (aiused) {
		printf("AI was used, score has\nbeen ignored.\n\n");
	}
	GKeyFlush();		//Clear the keyboard queue
	printf("Press [ESC] to exit\n");
	while (ngetchx() != KEY_ESC) {}
	
	Archive();
	exitquick:
	
	//Free the screen buffers, point arrays, and rotated sprites

	free(lbuffer);
	free(dbuffer);

	free(body);
	free(sprites);
	FontSetSys(font);
}

void NewFood(void) {
	pickpos:
	Food.x = random(20);
	Food.y = random(12);
	//This for loop checks to see if the new food position is on the snake's body, if it is it picks a new one.
	for (i = 1; i <= length; i++) {
		if (body[i].x == Food.x && body[i].y == Food.y) goto pickpos;
	}
}
void DrawSnake(void) {
	int sprt, vt;
for (i = 1; i <= length; i++) {

		sprt = 0;
		vt = 0;
		if (i == 1) {
			if (dir == 0)
				sprt = headright;
			else if (dir == 1)
				sprt = headdown;
			else if (dir == 2)
				sprt = headleft;
			else if (dir == 3)
				sprt = headup;
		}
		else if ((i == length) && (length > 1)) {
			if (IsAbove(i, i-1))
				sprt = tailup;
			else if (IsBelow(i, i-1))
				sprt = taildown;
			else if (IsRight(i, i-1))
				sprt = tailright;
			else if (IsLeft(i, i-1))
				sprt = tailleft;
		}
		else {
			if (IsBelow(i, i-1) || IsBelow(i, i+1)) {
				if (IsRight(i, i-1) || IsRight(i, i+1))
					sprt = upleft;
				else if (IsLeft(i, i-1) || IsLeft(i, i+1))
					sprt = upright;
				else
					sprt = bodyvert;
					vt++;
			}
			else if (IsAbove(i, i-1) || IsAbove(i, i+1)) {
				if (IsLeft(i, i-1) || IsLeft(i, i+1))
					sprt = downright;
				else
					sprt = downleft;
			}
			if (!sprt && !vt)
				sprt = bodyhorz;
		}
		GraySprite(body[i].x*8, body[i].y*8, sprt);
	}
}

void MoveSnake(void) {
	int gotfood = 0;
	if (body[1].x <= -1 || body[1].x >= 20 || body[1].y <= -1 || body[1].y >= 12) exit(0);		//See if you're hitting the wall
	//Check if you've eaten the food, if so, pick a new food position, increase score, and increase length by one
	if (body[1].x == Food.x && body[1].y == Food.y) {
		NewFood();
		length++;
		body[length].x = body[length - 1].x;
		body[length].y = body[length - 1].y;
		score = score + fps;
		gotfood = 1;
	}	
	//Check if you're trying to eat your tail
	for (i = 2; i <= length; i++) {
		if (body[1].x == body[i].x && body[1].y == body[i].y && gotfood == 0) exit(0);
	}
	//This for loop moves the data in the arrays back one spot to make room for the new head position
	for (i = length; i >= 2; i--) {
		body[i].x = body[i - 1].x;
		body[i].y = body[i - 1].y;
	}	
	//Depending on the direction, decide new head position
	if (dir == 0) body[1].x++;
	if (dir == 1) body[1].y++;
	if (dir == 2) body[1].x--;
	if (dir == 3) body[1].y--;
}


//Sprite editor!!!

#ifdef EDITOR
void SpriteEdit(void) {
	unsigned char cursor[] = {
		0b11100111,
		0b10000001,
		0b10000001,
		0b00000000,
		0b00000000,
		0b10000001,
		0b10000001,
		0b11100111
	};
	struct CURSOR {
		int x, y, ox, oy;
	} cur;
	sprtstart:
	clrscr();
	printf("What would you like to do?\n\n1. Edit sprites\n2. Save sprites to file\n3. Reset sprites\n\n");
	printf("Please press 1-3");
	key = ngetchx();
	int sc = 1,/* hl = 0, vl = 0, fl = 0,*/ sk = 0, sprt;
	memset(lbuffer, 0, LCD_SIZE);		//Clear light plane buffer
	memset(dbuffer, 0, LCD_SIZE);		//Clear dark plane buffer
	if (key == '1') {
		clrscr();
		GrayOn();
		cur.x = 0;
		cur.y = 0;
		cur.ox = -1;
		cur.oy = -1;
		sprt = headup;
		while (key != KEY_ESC) {
			if (cur.ox != -1) {
				Sprite8(cur.ox * 8 + 48, cur.oy * 8 + 18, 8, cursor, lbuffer, SPRT_XOR);
				Sprite8(cur.ox * 8 + 48, cur.oy * 8 + 18, 8, cursor, dbuffer, SPRT_XOR);
			}
			if (cur.ox == -1 || sc) {
				memset(lbuffer, 0, LCD_SIZE);		//Clear light plane buffer
				memset(dbuffer, 0, LCD_SIZE);		//Clear dark plane buffer
				RotateSprites();
				ScaleSprite8_OR(&sprites[sprt], lbuffer, 48, 18, 64, 64);
				ScaleSprite8_OR(&sprites[sprt+8], dbuffer, 48, 18, 64, 64);
				GraySprite(8, 8, downright);
				GraySprite(8, 16, bodyvert);
				GraySprite(8, 24, taildown);
				GraySprite(8, 40, food);
				GraySprite(16, 8, bodyhorz);
				GraySprite(24, 8, headright);
				sc = 0;
				GraySetAMSPlane(GrayGetPlane(DARK_PLANE));
			}
			Sprite8(cur.x * 8 + 48, cur.y * 8 + 18, 8, cursor, lbuffer, SPRT_XOR);
			Sprite8(cur.x * 8 + 48, cur.y * 8 + 18, 8, cursor, dbuffer, SPRT_XOR);
			FastCopyScreen(lbuffer, GrayGetPlane(LIGHT_PLANE));
			FastCopyScreen(dbuffer, GrayGetPlane(DARK_PLANE));
			cur.ox = cur.x;
			cur.oy = cur.y;
			key = ngetchx();
			if (key == KEY_UP && cur.y > 0) cur.y--;
			else if (key == KEY_DOWN && cur.y < 7) cur.y++;
			else if (key == KEY_LEFT && cur.x > 0) cur.x--;
			else if (key == KEY_RIGHT && cur.x < 7) cur.x++;
			
			if (key == '^')
			{
				DrawStrCenter("Copy to which sprite?", 0);
				sk = ngetchx();
				if (sk >= '1' && sk <= '5') {
					memcpy(&sprites[(sk-'1')*16], &sprites[sprt], 16);
					sc++;
				}
			}
			if (key == 'x') {
				memcpy(&sprites[sprt], &sprtdata[sprt], 16);
				sc++;
			}
			if (key >= '1' && key <= '5') {
				sprt = (key-'1')*16;
				sc++;
			}
			if (key == '+') { //horizontal line
				DrawStrCenter("Horz line, which color?", 0);
				key = ngetchx();
				if (key == KEY_F1) {
					for (i = 0; i <= 7; i++) {
						BitOff(sprites[sprt+i], cur.x);
						BitOff(sprites[sprt+i+8], cur.x);
						sc++;
					}
				}
				else if (key == KEY_F2) {
					for (i = 0; i <= 7; i++) {
						BitOn(sprites[sprt+i], cur.x);
						BitOff(sprites[sprt+i+8], cur.x);
						sc++;
					}
				}
				else if (key == KEY_F3) {
					for (i = 0; i <= 7; i++) {
						BitOff(sprites[sprt+i], cur.x);
						BitOn(sprites[sprt+i+8], cur.x);
						sc++;
					}
				}
				else if (key == KEY_F4) {
					for (i = 0; i <= 7; i++) {
						BitOn(sprites[sprt+i], cur.x);
						BitOn(sprites[sprt+i+8], cur.x);
						sc++;
					}
				}
			}
			else if (key == '-') { //vertical line
				DrawStrCenter("Vert line, which color?", 0);
				key = ngetchx();
				if (key == KEY_F1) {
					sprites[sprt+cur.y] = 0;
					sprites[sprt+cur.y+8] = 0;
					sc++;
				}
				else if (key == KEY_F2) {
					sprites[sprt+cur.y] = 255;
					sprites[sprt+cur.y+8] = 0;
					sc++;
				}
				else if (key == KEY_F3) {
					sprites[sprt+cur.y] = 0;
					sprites[sprt+cur.y+8] = 255;
					sc++;
				}
				else if (key == KEY_F4) {
					sprites[sprt+cur.y] = 255;
					sprites[sprt+cur.y+8] = 255;
					sc++;
				}
			}
			else if (key == '*') { //fill sprite with color
				DrawStrCenter("Fill sprite, which color?", 0);
				key = ngetchx();
				if (key == KEY_F1) {
					memset(&sprites[sprt], 0, 16);
					sc++;
				}
				else if (key == KEY_F2) {
					memset(&sprites[sprt], 255, 8);
					memset(&sprites[sprt+8], 0, 8);
					sc++;
				}
				else if (key == KEY_F3) {
					memset(&sprites[sprt], 0, 8);
					memset(&sprites[sprt+8], 255, 8);
					sc++;
				}
				else if (key == KEY_F4) {
					memset(&sprites[sprt], 255, 16);
					sc++;
				}
			}
			else if (key == '/') { //invert
				for (i = 0; i <= 15; i++)
					sprites[sprt+i] = ~(sprites[sprt+i]);
				sc++;
			}
			else if (key == KEY_F1) {
				BitOff(sprites[sprt+cur.y], cur.x);
				BitOff(sprites[sprt+cur.y+8], cur.x);
				sc++;
			}
			
			else if (key == KEY_F2) {
				BitOn(sprites[sprt+cur.y], cur.x);
				BitOff(sprites[sprt+cur.y+8], cur.x);
				sc++;
			}
			else if (key == KEY_F3) {
				BitOff(sprites[sprt+cur.y], cur.x);
				BitOn(sprites[sprt+cur.y+8], cur.x);
				sc++;
			}
			else if (key == KEY_F4) {
				BitOn(sprites[sprt+cur.y], cur.x);
				BitOn(sprites[sprt+cur.y+8], cur.x);
				sc++;
			}
		}
		GrayOff();
		key = 0;
	}
	else if (key == '2') {
		UnArchive();
		cfgfile = fopen("bnibcfg", "r+b");
		fseek(cfgfile, 3, SEEK_CUR);
		fwrite(sprites, 80, 1, cfgfile);
		fclose(cfgfile);
		Archive();
		clrscr();
		printf("Sprites saved, press any\nkey to return to the home\nscreen.");
		ngetchx();
		key = KEY_ESC;
	}
	else if (key == '3') {
		memcpy(sprites, sprtdata, 80);
		RotateSprites();
		clrscr();
		printf("Sprites reset.");
		ngetchx();
	}
	
	if (key != KEY_ESC) goto sprtstart;	
}
#endif


//Where the fun begins
void _main(void) {
	unsigned char interval;
	int odir;
	static int qdir;
	void *kbq = kbd_queue();
	char cbuffer[50];
  font = FontGetSys();
	FontSetSys(F_6x8);
	homequit = 0;
	//Allocate the screen buffers

	lbuffer = malloc(LCD_SIZE);
	dbuffer = malloc(LCD_SIZE);
	
	sprites = malloc(240);
	//set initial sprites
	memcpy(sprites, sprtdata, 80);
	
	//lbuffer = GrayGetPlane(LIGHT_PLANE);
	//dbuffer = GrayGetPlane(DARK_PLANE);
	
	atexit(ExitFunc);

	body = malloc(250*sizeof(struct BODY));
	//Assign initial values to some variables and randomize the generator...
	body[1].x = 2;
	body[1].y = 2;
	length = 1;
	dir = 0;
	odir = 0;
	qdir = 0;
	tmr = 1;
	ai = 0;
	aiused = 0;

	randomize();
	//Load the timer delay from the config file

	fps = 10;
	cfgfile = NULL;
	cfgfile = fopen("bnibcfg", "rb");
	if (cfgfile) {
		fread(&fps, 1, 1, cfgfile);
		fseek(cfgfile, 2, SEEK_CUR);
		fread(sprites, 80, 1, cfgfile);
		fclose(cfgfile);
	}
	
	//If config/sprite file isn't there, make it
	else {
		char fileex[] = {0, 'b', 'n', 'i', 'b', 0, OTH_TAG};
		cfgfile = fopen("bnibcfg", "wb");
		fputc(10,cfgfile);		//speed
		fputc(0,cfgfile);		//high score byte 1
		fputc(0,cfgfile);		//high score byte 2
		fwrite(sprtdata, 80, 1, cfgfile);
		fwrite(fileex, 7, 1, cfgfile);
		fclose(cfgfile);
		Archive();
		
	}
	
	//Make rotated sprites
	RotateSprites();
	
	//Display intro screen with speed selection
	homescr:
	clrscr();
	DrawStrCenter("Bnibbles v1.2.1", 0);
	DrawStrCenter("by Brett Simmers", 8);
	if (fps <= 30) sprintf(cbuffer, "Current speed is %d,", fps);
	else strcpy(cbuffer, "Current speed is MAX");
	DrawStrCenter(cbuffer, 24);
	DrawStrCenter("Hit [UP]/[DOWN] to adjust", 32);
	DrawStrCenter("Press [ESC] to exit,", 48);
	DrawStrCenter("any other key to start", 56);
	sprintf(cbuffer, "AI's state is: %d", ai);
	DrawStrCenter(cbuffer, 72);
	DrawStrCenter("Press [MODE] to toggle", 80);

	#ifdef EDITOR
	DrawStrCenter("[APPS] - Sprite Editor", 88);
	#endif
	
	key = ngetchx();
	if (key == KEY_ESC) {
		homequit = 1;
		exit(0);
	}
	if (key == KEY_UP) {
		if (fps < 31) fps++;
		goto homescr;
	}
	if (key == KEY_DOWN) {
		if (fps > 5) fps--;
		goto homescr;
	}
	if (key == KEY_MODE) {
		if (ai == 1) ai = 0;
		else ai = 1;
		goto homescr;
	}
	#ifdef EDITOR
	if (key == KEY_APPS) {
		SpriteEdit();
		goto homescr;
	}
	#endif
	interval = 257 - (1060 / (fps+1));
	pokeIO(0x600017, interval);
	clrscr();

	GrayOn();

	NewFood();		//Pick a random food position
	//Set up interrupt handlers
	OldInt5 = GetIntVec(AUTO_INT_5);
	SetIntVec(AUTO_INT_5, CustomInt5);
	

	
	score = fps;
	//Main Loop
	doit:
	if (fps != 31) tmr = 0;
	
	//AI----------------------------
	if (ai == 1) {
		//GET THE FOOD!!!!
		if ((body[1].x == Food.x) && (body[1].y > Food.y) && (odir != 1)) dir = 3;
		if ((body[1].x == Food.x) && (body[1].y < Food.y) && (odir != 3)) dir = 1;
		if ((body[1].y == Food.y) && (body[1].x > Food.x) && (odir != 0)) dir = 2;
		if ((body[1].y == Food.y) && (body[1].x < Food.x) && (odir != 2)) dir = 0;
		
		//don't eat yourself!!!
		if (dir == 0) {
			for (i = 2; i <= length; i++) {
				if (IsLeft(1, i)) dir = 3;
			}
		}
		if (dir == 1) {
			for (i = 2; i <= length; i++) {
				if (IsAbove(1, i)) dir = 0;
			}
		}
		if (dir == 2) {
			for (i = 2; i <= length; i++) {
				if (IsRight(1, i)) dir = 1;
			}
		}
		if (dir == 3) {
			for (i = 2; i <= length; i++) {
				if (IsBelow(1, i)) dir = 2;
			}
		}
		//Keep away from the walls!!!
		if ((body[1].x == 19) && (odir == 0)) dir = 1;
		if ((body[1].x == 0) && (odir == 2)) dir = 3;
		if ((body[1].y == 11) && (odir == 1)) dir = 2;
		if ((body[1].y == 0) && (odir == 3)) dir = 0;
		if (_keytest(RR_ESC)) exit(0);
		
		aiused = 1;   //This is just so that you can't get a high score if the ai was on at any point during the run
	}
	//----------------------------End of ai
	while (tmr == 0) {
		if (!OSdequeue(&key, kbq)) {
			if (key == KEY_ESC) exit(0);
			if (key == KEY_RIGHT && dir != 2) {
				if (odir == 2) qdir = 4;
				else dir = 0;
			}
			if (key == KEY_DOWN && dir != 3) {
				if (odir == 3) qdir = 5;
				else dir = 1;
			}
			if (key == KEY_LEFT && dir != 0) {
				if (odir == 0) qdir = 6;
				else dir = 2;
			}
			if (key == KEY_UP && dir != 1) {
				if (odir == 1) qdir = 7;
				else dir = 3;
			}
			if (key == 16721) {
				if (fps < 30) fps++;
				interval = 257 - (1060 / (fps+1));
				pokeIO(0x600017, interval);
			}
			if (key == 16724) {
				if (fps > 5) fps--;
				interval = 257 - (1060 / (fps+1));
				pokeIO(0x600017, interval);
			}
			if (key == KEY_MODE) {
				if (ai == 1) ai = 0;
				else ai = 1;
			}
		}
	}
	if ((qdir != 0) && (qdir < 4)) {
		dir = qdir;
		qdir = 0;
	}
	if (qdir > 3) qdir = qdir - 4;
	odir = dir;
	
	MoveSnake();

	memset(lbuffer, 0, LCD_SIZE);		//Clear light plane buffer
	memset(dbuffer, 0, LCD_SIZE);		//Clear dark plane buffer
	memset(lbuffer+0xb40, 0xFF, 120);
	memset(dbuffer+0xb40,0xFF, 120);
	DrawSnake();
	
	//Draw food
	GraySprite(Food.x*8, Food.y*8, food);
	
	FastCopyScreen(lbuffer, GrayGetPlane(LIGHT_PLANE));
	FastCopyScreen(dbuffer, GrayGetPlane(DARK_PLANE));
	

	

	goto doit;
}
//1060/x, x is times per second
